# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
# Copyright (C) 2024 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Admin views for managing access to actions."""

from flask_admin.contrib.sqla import ModelView
from invenio_db import db

from .models import Client, Token


def _(x):
    """Identity."""
    return x


class ClientView(ModelView):
    """View for managing access to actions by users."""

    can_delete = True
    can_create = False
    can_view_details = True
    column_display_all_relations = True

    list_all = (
        "name",
        "description",
        "website",
        "user_id",
        "client_id",
    )

    column_list = column_searchable_list = column_sortable_list = (
        column_details_list
    ) = list_all

    column_list = list_all
    column_default_sort = ("client_id", True)


class TokenView(ModelView):
    """View for managing access to actions by users with given role."""

    can_delete = True
    can_create = False
    can_view_details = True
    list_all = (
        "id",
        "client_id",
        "user_id",
        "token_type",
        "expires",
    )
    column_list = column_searchable_list = column_sortable_list = (
        column_details_list
    ) = list_all


oauth2server_clients_adminview = {
    "view_class": ClientView,
    "args": [Client, db.session],
    "kwargs": {
        "name": _("OAuth Applications"),
        "category": _("User Management"),
    },
}

oauth2server_tokens_adminview = {
    "view_class": TokenView,
    "args": [Token, db.session],
    "kwargs": {
        "name": _("OAuth Application Tokens"),
        "category": _("User Management"),
    },
}

__all__ = (
    "oauth2server_clients_adminview",
    "oauth2server_tokens_adminview",
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
# Copyright (C) 2024 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create oauth2server tables."""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op

# revision identifiers, used by Alembic.
revision = "12a88921ada2"
down_revision = "aa546f2a8d2f"
branch_labels = ()
depends_on = "9848d0149abd"


def upgrade():
    """Upgrade database."""
    op.create_table(
        "oauth2server_client",
        sa.Column("name", sa.String(length=40), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("website", sqlalchemy_utils.types.url.URLType(), nullable=True),
        sa.Column("user_id", sa.Integer(), nullable=True),
        sa.Column("client_id", sa.String(length=255), nullable=False),
        sa.Column("client_secret", sa.String(length=255), nullable=False),
        sa.Column("is_confidential", sa.Boolean(name="is_confidential"), nullable=True),
        sa.Column("is_internal", sa.Boolean(name="is_internal"), nullable=True),
        sa.Column("_redirect_uris", sa.Text(), nullable=True),
        sa.Column("_default_scopes", sa.Text(), nullable=True),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["accounts_user.id"],
        ),
        sa.PrimaryKeyConstraint("client_id"),
    )
    op.create_index(
        op.f("ix_oauth2server_client_client_secret"),
        "oauth2server_client",
        ["client_secret"],
        unique=True,
    )
    op.create_table(
        "oauth2server_token",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("client_id", sa.String(length=255), nullable=False),
        sa.Column("user_id", sa.Integer(), nullable=True),
        sa.Column("token_type", sa.String(length=255), nullable=True),
        sa.Column(
            "access_token", sqlalchemy_utils.StringEncryptedType(), nullable=True
        ),
        sa.Column(
            "refresh_token", sqlalchemy_utils.StringEncryptedType(), nullable=True
        ),
        sa.Column("expires", sa.DateTime(), nullable=True),
        sa.Column("_scopes", sa.Text(), nullable=True),
        sa.Column("is_personal", sa.Boolean(name="is_personal"), nullable=True),
        sa.Column("is_internal", sa.Boolean(name="is_internal"), nullable=True),
        sa.ForeignKeyConstraint(
            ["client_id"],
            ["oauth2server_client.client_id"],
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["accounts_user.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        "ix_oauth2server_token_access_token",
        "oauth2server_token",
        ["access_token"],
        unique=True,
        mysql_length=255,
    )
    op.create_index(
        "ix_oauth2server_token_refresh_token",
        "oauth2server_token",
        ["refresh_token"],
        unique=True,
        mysql_length=255,
    )


def downgrade():
    """Downgrade database."""
    op.drop_index(
        "ix_oauth2server_token_refresh_token", table_name="oauth2server_token"
    )
    op.drop_index("ix_oauth2server_token_access_token", table_name="oauth2server_token")
    op.drop_table("oauth2server_token")
    op.drop_index(
        op.f("ix_oauth2server_client_client_secret"), table_name="oauth2server_client"
    )
    op.drop_table("oauth2server_client")



# This file is part of Invenio.
# Copyright (C) 2015-2020 CERN.
# Copyright (C) 2024 Graz University of Technology.
# Copyright (C) 2024 KTH Royal Institute of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Render semantic ui styling for forms."""

from markupsafe import Markup


class SelectSUI(object):
    """Renders a select field with Semantic UI styling."""

    def __init__(self, multiple=False):
        """Init."""
        self.multiple = multiple

    def __call__(self, field, **kwargs):
        """Render select field."""
        html = [
            '<div role="listbox" aria-expanded="false" class="ui fluid selection dropdown">'
        ]
        html.append('<input type="hidden" name="{0}">'.format(field.name))
        html.append('<i class="dropdown icon" aria-hidden="true"></i>')
        items_html = []
        default_text_html = []
        for val, label, selected, _ in field.iter_choices():
            if selected:
                default_text_html = [
                    '<div aria-atomic="true" aria-live="polite" class="text">{0}</div><div class="menu">'.format(
                        label
                    )
                ]
            items_html.append(self.render_option(val, label, selected))
        items_html.append("</div></div>")
        html = html + default_text_html + items_html
        return Markup("".join(html))

    @classmethod
    def render_option(cls, value, label, selected, **kwargs):
        """Render option."""
        return Markup(
            (
                '<div role="option" aria-selected="{2}" class="item" data-value="{0}">{1}</div>'
            ).format(value, label, selected)
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test invenio_oauth2server validators."""

import pytest
from oauthlib.oauth2.rfc6749.errors import (
    InsecureTransportError,
    InvalidRedirectURIError,
)
from wtforms.validators import ValidationError

from invenio_oauth2server.validators import URLValidator, validate_redirect_uri


@pytest.mark.parametrize(
    "input,expected",
    [
        ("example.org/", InvalidRedirectURIError()),
        ("http://", InvalidRedirectURIError()),
        ("http://example.org/", InsecureTransportError()),
        ("https://example.org/", None),
        ("https://localhost/", None),
        ("https://127.0.0.1", None),
    ],
)
def test_validate_redirect_uri(input, expected):
    """Test redirect URI validator."""
    try:
        validate_redirect_uri(input)
    except Exception as e:
        assert type(e) is type(expected)


def test_url_validator(app):
    """Test url validator."""

    class Field(object):
        def __init__(self, data):
            self.data = data

        def gettext(self, *args, **kwargs):
            return "text"

    with app.app_context():
        # usually if localhost, validation is failing
        with pytest.raises(ValidationError):
            URLValidator()(form=None, field=Field(data="http://localhost:5000"))
        URLValidator()(form=None, field=Field(data="http://mywebsite.it:5000"))

        # enable debug mode to accept also localhost
        app.config["DEBUG"] = True
        URLValidator()(form=None, field=Field(data="http://localhost:5000"))
        URLValidator()(form=None, field=Field(data="http://mywebsite.it:5000"))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Add on delete cascade."""

from alembic import op

# revision identifiers, used by Alembic.
revision = "4e57407b8e4a"
down_revision = "12a88921ada2"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.drop_constraint(
        "fk_oauth2server_client_user_id_accounts_user",
        "oauth2server_client",
        type_="foreignkey",
    )
    op.create_foreign_key(
        op.f("fk_oauth2server_client_user_id_accounts_user"),
        "oauth2server_client",
        "accounts_user",
        ["user_id"],
        ["id"],
        ondelete="CASCADE",
    )
    op.create_index(
        op.f("ix_oauth2server_client_user_id"),
        "oauth2server_client",
        ["user_id"],
        unique=False,
    )
    op.drop_constraint(
        "fk_oauth2server_token_user_id_accounts_user",
        "oauth2server_token",
        type_="foreignkey",
    )
    op.create_foreign_key(
        op.f("fk_oauth2server_token_user_id_accounts_user"),
        "oauth2server_token",
        "accounts_user",
        ["user_id"],
        ["id"],
        ondelete="CASCADE",
    )
    op.create_index(
        op.f("ix_oauth2server_token_user_id"),
        "oauth2server_token",
        ["user_id"],
        unique=False,
    )
    op.drop_constraint(
        "fk_oauth2server_token_client_id_oauth2server_client",
        "oauth2server_token",
        type_="foreignkey",
    )
    op.create_foreign_key(
        op.f("fk_oauth2server_token_client_id_oauth2server_client"),
        "oauth2server_token",
        "oauth2server_client",
        ["client_id"],
        ["client_id"],
        ondelete="CASCADE",
    )
    op.create_index(
        op.f("ix_oauth2server_token_client_id"),
        "oauth2server_token",
        ["client_id"],
        unique=False,
    )


def downgrade():
    """Downgrade database."""
    op.drop_constraint(
        op.f("fk_oauth2server_token_user_id_accounts_user"),
        "oauth2server_token",
        type_="foreignkey",
    )
    op.drop_index(
        op.f("ix_oauth2server_token_user_id"), table_name="oauth2server_token"
    )
    op.create_foreign_key(
        "fk_oauth2server_token_user_id_accounts_user",
        "oauth2server_token",
        "accounts_user",
        ["user_id"],
        ["id"],
    )
    op.drop_constraint(
        op.f("fk_oauth2server_token_client_id_oauth2server_client"),
        "oauth2server_token",
        type_="foreignkey",
    )
    op.drop_index(
        op.f("ix_oauth2server_token_client_id"), table_name="oauth2server_token"
    )
    op.create_foreign_key(
        "fk_oauth2server_token_client_id_oauth2server_client",
        "oauth2server_token",
        "oauth2server_client",
        ["client_id"],
        ["client_id"],
    )
    op.drop_constraint(
        op.f("fk_oauth2server_client_user_id_accounts_user"),
        "oauth2server_client",
        type_="foreignkey",
    )
    op.drop_index(
        op.f("ix_oauth2server_client_user_id"), table_name="oauth2server_client"
    )
    op.create_foreign_key(
        "fk_oauth2server_client_user_id_accounts_user",
        "oauth2server_client",
        "accounts_user",
        ["user_id"],
        ["id"],
    )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2024 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Define forms for generating access tokens and clients."""

from flask import current_app
from flask_wtf import FlaskForm as Form
from invenio_i18n import lazy_gettext as _
from markupsafe import Markup
from oauthlib.oauth2.rfc6749.errors import (
    InsecureTransportError,
    InvalidRedirectURIError,
)
from werkzeug.local import LocalProxy
from wtforms import fields, validators, widgets
from wtforms_alchemy import model_form_factory

from .models import Client
from .theme.semantic.form_styling import SelectSUI
from .validators import URLValidator, validate_redirect_uri


#
# Widget
#
def scopes_multi_checkbox(field, **kwargs):
    """Render multi checkbox widget."""
    kwargs.setdefault("type", "checkbox")
    field_id = kwargs.pop("id", field.id)

    html = ['<div class="row">']

    for value, label, checked, render_kw in field.iter_choices():
        choice_id = "%s-%s" % (field_id, value)

        options = dict(
            kwargs,
            name=field.name,
            value=value,
            id=choice_id,
            class_=" ",
        )

        if checked:
            options["checked"] = "checked"

        html.append('<div class="col-md-3">')
        html.append('<label for="{0}" class="checkbox-inline">'.format(choice_id))
        html.append("<input {0} /> ".format(widgets.html_params(**options)))
        html.append(
            '{0} <br/><small class="text-muted">{1}</small>'.format(
                value, label.help_text
            )
        )
        html.append("</label></div>")
    html.append("</div>")

    return Markup("".join(html))


#
# Redirect URI field
#
class RedirectURIField(fields.TextAreaField):
    """Process redirect URI field data."""

    def process_formdata(self, valuelist):
        """Process form data."""
        if valuelist:
            self.data = "\n".join(
                [
                    x.strip()
                    for x in filter(lambda x: x, "\n".join(valuelist).splitlines())
                ]
            )

    def process_data(self, value):
        """Process data."""
        self.data = "\n".join(value)


class RedirectURIValidator(object):
    """Validate if redirect URIs."""

    def __call__(self, form, field):
        """Call function."""
        errors = []
        for v in field.data.splitlines():
            try:
                validate_redirect_uri(v)
            except InsecureTransportError:
                errors.append(v)
            except InvalidRedirectURIError:
                errors.append(v)

        if errors:
            raise validators.ValidationError(
                _("Invalid redirect URIs: {urls}").format(urls=", ".join(errors))
            )


#
# Forms
#
class ClientFormBase(model_form_factory(Form)):
    """Base class for Client form."""

    class Meta:
        """Metadata class."""

        model = Client
        exclude = [
            "client_secret",
            "is_internal",
        ]
        strip_string_fields = True
        field_args = dict(
            website=dict(
                validators=[validators.DataRequired(), URLValidator()],
                widget=widgets.TextInput(),
            ),
        )


class ClientForm(ClientFormBase):
    """Client form."""

    # Trick to make redirect_uris render in the bottom of the form.
    redirect_uris = RedirectURIField(
        label=_("Redirect URIs (one per line)"),
        description=_(
            "One redirect URI per line. This is your application's"
            " authorization callback URLs. HTTPS must be used for all "
            "hosts except localhost (for testing purposes)."
        ),
        validators=[RedirectURIValidator(), validators.DataRequired()],
        default="",
    )

    widget = LocalProxy(
        lambda: (
            SelectSUI()
            if current_app.config.get("APP_THEME")
            and current_app.config.get("APP_THEME")[0] == "semantic-ui"
            else widgets.Select()
        )
    )

    is_confidential = fields.SelectField(
        label=_("Client type"),
        description=_(
            "Select confidential if your application is capable of keeping "
            "the issued client secret confidential (e.g. a web application), "
            "select public if your application cannot (e.g. a browser-based "
            "JavaScript application). If you select public, your application "
            "MUST validate the redirect URI."
        ),
        coerce=lambda x: False if x == "False" else bool(x),
        choices=[("True", _("Confidential")), ("False", _("Public"))],
        default="True",
        widget=widget,
    )


class TokenForm(Form):
    """Token form."""

    name = fields.StringField(
        description=_("Name of personal access token."),
        validators=[
            validators.DataRequired(),
            validators.Length(
                max=40, message=_("The name must be less than 40 characters long.")
            ),
        ],
    )

    scopes = fields.SelectMultipleField(
        widget=scopes_multi_checkbox,
        choices=[],  # Must be dynamically provided in view.
        description=_(
            "Scopes assign permissions to your personal access token."
            " A personal access token works just like a normal OAuth "
            " access token for authentication against the API."
        ),
    )
